@@ -1,24 +1,25 @@
-import json
 import os
 import pathlib
 import signal
+import sys
 import traceback
-from typing import List, Optional, Tuple
+from typing import List, Optional, Tuple, Dict
 
 import typer
 from rich import print as rprint
 
-from skyplane import exceptions, GB, format_bytes, gateway_docker_image, skyplane_root
+from skyplane import exceptions, GB, format_bytes, gateway_docker_image, skyplane_root, cloud_config
 from skyplane.compute.cloud_providers import CloudProvider
 from skyplane.obj_store.object_store_interface import ObjectStoreInterface, ObjectStoreObject
 from skyplane.obj_store.s3_interface import S3Object
 from skyplane.obj_store.gcs_interface import GCSObject
 from skyplane.obj_store.azure_blob_interface import AzureBlobObject
 from skyplane.replicate.replication_plan import ReplicationTopology, ReplicationJob
-from skyplane.replicate.replicator_client import ReplicatorClient
+from skyplane.replicate.replicator_client import ReplicatorClient, TransferStats
 from skyplane.utils import logger
 from skyplane.utils.timer import Timer
 from skyplane.cli.common import console
+from skyplane.cli.usage.client import UsageClient
 
 
 def generate_topology(
@@ -27,22 +28,29 @@ def generate_topology(
     solve: bool,
     num_connections: int = 32,
     max_instances: int = 1,
+    solver_class: str = "ILP",
     solver_total_gbyte_to_transfer: Optional[float] = None,
     solver_required_throughput_gbits: float = 4,
     solver_throughput_grid: Optional[pathlib.Path] = skyplane_root / "profiles" / "throughput.csv",
     solver_verbose: Optional[bool] = False,
+    args: Optional[Dict] = None,
 ) -> ReplicationTopology:
-    if solve:
-        if src_region == dst_region:
-            typer.secho("Solver is not supported for intra-region transfers, run without the --solve flag", fg="red")
-            raise typer.Exit(1)
-
-        # build problem and solve
+    if src_region == dst_region:  # intra-region transfer w/o solver
+        topo = ReplicationTopology()
+        for i in range(max_instances):
+            topo.add_objstore_instance_edge(src_region, src_region, i)
+            topo.add_instance_objstore_edge(src_region, i, src_region)
+        topo.cost_per_gb = 0
+        return topo
+    elif solve:
         from skyplane.replicate.solver import ThroughputProblem
-        from skyplane.replicate.solver_ilp import ThroughputSolverILP
 
+        if src_region == dst_region:
+            e = "Solver is not supported for intra-region transfers, run without the --solve flag"
+            typer.secho(e, fg="red", err=True)
+            UsageClient.log_exception("generate_topology", exceptions.SkyplaneException(e), args, src_region, dst_region)
+            raise typer.Exit(1)
         assert solver_throughput_grid is not None and solver_total_gbyte_to_transfer is not None
-        tput = ThroughputSolverILP(solver_throughput_grid)
         problem = ThroughputProblem(
             src=src_region,
             dst=dst_region,
@@ -50,63 +58,80 @@ def generate_topology(
             gbyte_to_transfer=solver_total_gbyte_to_transfer,
             instance_limit=max_instances,
         )
-        with Timer() as t:
-            with console.status("Solving for the optimal transfer plan"):
-                solution = tput.solve_min_cost(
-                    problem, solver=ThroughputSolverILP.choose_solver(), solver_verbose=solver_verbose, save_lp_path=None
-                )
-        typer.secho(f"Solving for the optimal transfer plan took {t.elapsed:.2f}s", fg="green")
-        topo, scale_factor = tput.to_replication_topology(solution)
-        logger.fs.debug(f"Scaled solution by {scale_factor:.2f}x")
-        topo.cost_per_gb = solution.cost_egress / solution.problem.gbyte_to_transfer
-        return topo
-    else:
-        if src_region == dst_region:
-            topo = ReplicationTopology()
-            for i in range(max_instances):
-                topo.add_objstore_instance_edge(src_region, src_region, i)
-                topo.add_instance_objstore_edge(src_region, i, src_region)
-            topo.cost_per_gb = 0
+
+        if solver_class == "ILP":
+            from skyplane.replicate.solver_ilp import ThroughputSolverILP
+
+            tput = ThroughputSolverILP(solver_throughput_grid)
+            with Timer() as t:
+                with console.status("Solving for the optimal transfer plan"):
+                    solution = tput.solve_min_cost(
+                        problem, solver=ThroughputSolverILP.choose_solver(), solver_verbose=solver_verbose, save_lp_path=None
+                    )
+            typer.secho(f"Solving for the optimal transfer plan took {t.elapsed:.2f}s", fg="green")
+            topo, scale_factor = tput.to_replication_topology(solution)
+            logger.fs.debug(f"Scaled solution by {scale_factor:.2f}x")
+            topo.cost_per_gb = solution.cost_egress / solution.problem.gbyte_to_transfer
+            return topo
+        elif solver_class == "RON":
+            from skyplane.replicate.solver_ron import ThroughputSolverRON
+
+            tput = ThroughputSolverRON(solver_throughput_grid)
+            solution = tput.solve(problem)
+            topo, scale_factor = tput.to_replication_topology(solution)
+            topo.cost_per_gb = solution.cost_egress / solution.problem.gbyte_to_transfer
+            return topo
         else:
-            topo = ReplicationTopology()
-            for i in range(max_instances):
-                topo.add_objstore_instance_edge(src_region, src_region, i)
-                topo.add_instance_instance_edge(src_region, i, dst_region, i, num_connections)
-                topo.add_instance_objstore_edge(dst_region, i, dst_region)
-            topo.cost_per_gb = CloudProvider.get_transfer_cost(src_region, dst_region)
+            raise NotImplementedError(f"Solver class {solver_class} not implemented")
+    else:  # inter-region transfer w/o solver
+        topo = ReplicationTopology()
+        for i in range(max_instances):
+            topo.add_objstore_instance_edge(src_region, src_region, i)
+            topo.add_instance_instance_edge(src_region, i, dst_region, i, num_connections)
+            topo.add_instance_objstore_edge(dst_region, i, dst_region)
+        topo.cost_per_gb = CloudProvider.get_transfer_cost(src_region, dst_region)
         return topo
 
 
-def map_object_key_prefix(
-    source_prefix: str,
-    source_key: str,
-    dest_prefix: str,
-    recursive: bool = False,
-):
+def map_object_key_prefix(source_prefix: str, source_key: str, dest_prefix: str, recursive: bool = False):
     """
     map_object_key_prefix computes the mapping of a source key in a bucket prefix to the destination.
     Users invoke a transfer via the CLI; aws s3 cp s3://bucket/source_prefix s3://bucket/dest_prefix.
     The CLI will query the object store for all objects in the source prefix and map them to the
     destination prefix using this function.
     """
+    join = lambda prefix, fname: prefix + fname if prefix.endswith("/") else prefix + "/" + fname
+    src_fname = source_key.split("/")[-1] if "/" in source_key and not source_key.endswith("/") else source_key
     if not recursive:
-        if source_key.startswith(source_prefix):
-            if dest_prefix.endswith("/"):
-                src_name = source_key.split("/")[-1]
-                return f"{dest_prefix}{src_name}"
-            elif source_prefix == source_key:
+        if source_key == source_prefix:
+            if dest_prefix == "" or dest_prefix == "/":
+                return src_fname
+            elif dest_prefix[-1] == "/":
+                return dest_prefix + src_fname
+            else:
                 return dest_prefix
-        raise exceptions.MissingObjectException(
-            f"Source key {source_key} does not start with source prefix {source_prefix}. To copy a directory, please use the '--recursive' flag"
-        )
-    else:
-        dest_prefix = dest_prefix if dest_prefix.endswith("/") else f"{dest_prefix}/"
-        source_prefix = source_prefix if source_prefix.endswith("/") else f"{source_prefix}/"
-        if source_key.startswith(source_prefix):
-            file_path = source_key[len(source_prefix) :]
-            return f"{dest_prefix}{file_path}"
         else:
-            raise exceptions.MissingObjectException(f"Source key {source_key} does not start with source prefix {source_prefix}")
+            rprint(f"\n:x: [bold red]In order to transfer objects using a prefix, you must use the --recursive or -r flag.[/bold red]")
+            rprint(f"[yellow]If you meant to transfer a single object, pass the full source object key.[/yellow]")
+            rprint(f"[bright_black]Try running: [bold]skyplane {' '.join(sys.argv[1:])} --recursive[/bold][/bright_black]")
+            raise exceptions.MissingObjectException("Encountered a recursive transfer without the --recursive flag.")
+    else:
+        if source_prefix == "" or source_prefix == "/":
+            if dest_prefix == "" or dest_prefix == "/":
+                return source_key
+            else:
+                return join(dest_prefix, source_key)
+        else:
+            # catch special case: map_object_key_prefix("foo", "foobar/baz.txt", "", recursive=True)
+            if not source_key.startswith(source_prefix + "/" if not source_prefix.endswith("/") else source_prefix):
+                rprint(f"\n:x: [bold red]The source key {source_key} does not start with the source prefix {source_prefix}[/bold red]")
+                raise exceptions.MissingObjectException(f"Source key {source_key} does not start with source prefix {source_prefix}")
+            if dest_prefix == "" or dest_prefix == "/":
+                return source_key[len(source_prefix) :]
+            else:
+                src_path_after_prefix = source_key[len(source_prefix) :]
+                src_path_after_prefix = src_path_after_prefix[1:] if src_path_after_prefix.startswith("/") else src_path_after_prefix
+                return join(dest_prefix, src_path_after_prefix)
 
 
 def generate_full_transferobjlist(
@@ -118,16 +143,19 @@ def generate_full_transferobjlist(
     dest_prefix: str,
     recursive: bool = False,
 ) -> List[Tuple[ObjectStoreObject, ObjectStoreObject]]:
-    """Query source region and destination region buckets and return list of objects to transfer."""
+    """Query source region and return list of objects to transfer."""
     source_iface = ObjectStoreInterface.create(source_region, source_bucket)
     dest_iface = ObjectStoreInterface.create(dest_region, dest_bucket)
 
+    requester_pays = cloud_config.get_flag("requester_pays")
+    if requester_pays:
+        source_iface.set_requester_bool(True)
+
     # ensure buckets exist
     if not source_iface.bucket_exists():
         raise exceptions.MissingBucketException(f"Source bucket {source_bucket} does not exist")
     if not dest_iface.bucket_exists():
         raise exceptions.MissingBucketException(f"Destination bucket {dest_bucket} does not exist")
-
     source_objs, dest_objs = [], []
 
     # query all source region objects
@@ -137,12 +165,15 @@ def generate_full_transferobjlist(
             source_objs.append(obj)
             status.update(f"Querying objects in {source_bucket} (found {len(source_objs)} objects so far)")
     if not source_objs:
-        logger.error("Specified object does not exist.")
+        logger.error("Specified object does not exist.\n")
         raise exceptions.MissingObjectException(f"No objects were found in the specified prefix {source_prefix} in {source_bucket}")
 
     # map objects to destination object paths
     for source_obj in source_objs:
-        dest_key = map_object_key_prefix(source_prefix, source_obj.key, dest_prefix, recursive=recursive)
+        try:
+            dest_key = map_object_key_prefix(source_prefix, source_obj.key, dest_prefix, recursive=recursive)
+        except exceptions.MissingObjectException:
+            raise typer.Exit(1)
         if dest_region.startswith("aws"):
             dest_obj = S3Object(dest_region.split(":")[0], dest_bucket, dest_key)
         elif dest_region.startswith("gcp"):
@@ -154,6 +185,15 @@ def generate_full_transferobjlist(
         # dest_obj = ObjectStoreObject(dest_region.split(":")[0], dest_bucket, dest_key)
         dest_objs.append(dest_obj)
 
+    return list(zip(source_objs, dest_objs))
+
+
+def enrich_dest_objs(dest_region: str, dest_prefix: str, dest_bucket: str, dest_objs: list):
+    """
+    For skyplane sync, we enrich dest obj metadata with our existing dest obj metadata from the dest bucket following a query.
+    """
+    dest_iface = ObjectStoreInterface.create(dest_region, dest_bucket)
+
     # query destination at dest_key
     logger.fs.debug(f"Querying objects in {dest_bucket}")
     dest_objs_keys = {obj.key for obj in dest_objs}
@@ -171,8 +211,6 @@ def generate_full_transferobjlist(
             dest_obj.size = found_dest_objs[dest_obj.key].size
             dest_obj.last_modified = found_dest_objs[dest_obj.key].last_modified
 
-    return list(zip(source_objs, dest_objs))
-
 
 def confirm_transfer(topo: ReplicationTopology, job: ReplicationJob, ask_to_confirm_transfer=True):
     console.print(
@@ -223,16 +261,20 @@ def launch_replication_job(
     # multipart
     multipart_enabled: bool = False,
     multipart_min_threshold_mb: int = 128,
-    multipart_min_size_mb: int = 8,
+    multipart_chunk_size_mb: int = 64,
     multipart_max_chunks: int = 9990,
     # cloud provider specific options
+    aws_use_spot_instances: bool = False,
     aws_instance_class: str = "m5.8xlarge",
+    azure_use_spot_instances: bool = False,
     azure_instance_class: str = "Standard_D32_v4",
+    gcp_use_spot_instances: bool = False,
     gcp_instance_class: str = "n2-standard-32",
     gcp_use_premium_network: bool = True,
     # logging options
     time_limit_seconds: Optional[int] = None,
     log_interval_s: float = 1.0,
+    error_reporting_args: Optional[Dict] = None,
 ):
     if "SKYPLANE_DOCKER_IMAGE" in os.environ:
         rprint(f"[bright_black]Using overridden docker image: {gateway_docker_image}[/bright_black]")
@@ -240,6 +282,7 @@ def launch_replication_job(
         typer.secho(
             f"Instances will remain up and may result in continued cloud billing. Remember to call `skyplane deprovision` to deprovision gateways.",
             fg="red",
+            err=True,
             bold=True,
         )
 
@@ -252,13 +295,20 @@ def launch_replication_job(
         gcp_instance_class=gcp_instance_class,
         gcp_use_premium_network=gcp_use_premium_network,
     )
-    typer.secho(f"Storing debug information for transfer in {rc.transfer_dir / 'client.log'}", fg="yellow")
+    typer.secho(f"Storing debug information for transfer in {rc.transfer_dir / 'client.log'}", fg="yellow", err=True)
     (rc.transfer_dir / "topology.json").write_text(topo.to_json())
 
-    stats = {}
+    stats = TransferStats.empty()
     try:
         rc.provision_gateways(
-            reuse_gateways, use_bbr=use_bbr, use_compression=use_compression, use_e2ee=use_e2ee, use_socket_tls=use_socket_tls
+            reuse_gateways,
+            use_bbr=use_bbr,
+            use_compression=use_compression,
+            use_e2ee=use_e2ee,
+            use_socket_tls=use_socket_tls,
+            aws_use_spot_instances=aws_use_spot_instances,
+            azure_use_spot_instances=azure_use_spot_instances,
+            gcp_use_spot_instances=gcp_use_spot_instances,
         )
         for node, gw in rc.bound_nodes.items():
             logger.fs.info(f"Log URLs for {gw.uuid()} ({node.region}:{node.instance})")
@@ -268,20 +318,17 @@ def launch_replication_job(
             job,
             multipart_enabled=multipart_enabled,
             multipart_min_threshold_mb=multipart_min_threshold_mb,
-            multipart_min_size_mb=multipart_min_size_mb,
+            multipart_chunk_size_mb=multipart_chunk_size_mb,
             multipart_max_chunks=multipart_max_chunks,
         )
         total_bytes = sum([chunk_req.chunk.chunk_length_bytes for chunk_req in job.chunk_requests])
         console.print(f":rocket: [bold blue]{total_bytes / GB:.2f}GB transfer job launched[/bold blue]")
-        if topo.source_region().split(":")[0] == "azure" or topo.sink_region().split(":")[0] == "azure":
-            typer.secho(
-                f"Warning: For Azure transfers, your transfer may block for up to 120s waiting for role assignments to propagate. See issue #355.",
-                fg="yellow",
-            )
+
         stats = rc.monitor_transfer(
             job,
             show_spinner=True,
             log_interval_s=log_interval_s,
+            log_to_file=True,
             time_limit_seconds=time_limit_seconds,
             multipart=multipart_enabled,
             write_profile=debug,
@@ -300,33 +347,35 @@ def launch_replication_job(
             s = signal.signal(signal.SIGINT, signal.SIG_IGN)
             rc.deprovision_gateways()
             signal.signal(signal.SIGINT, s)
-
-        stats["success"] = False
-        out_json = {k: v for k, v in stats.items() if k not in ["log", "completed_chunk_ids"]}
-        typer.echo(f"\n{json.dumps(out_json)}")
+        UsageClient.log_exception("launch_replication_job", e, error_reporting_args, job.source_region, job.dest_region)
         os._exit(1)  # exit now
+
     if not reuse_gateways:
         s = signal.signal(signal.SIGINT, signal.SIG_IGN)
         rc.deprovision_gateways()
         signal.signal(signal.SIGINT, s)
-
-    stats = stats if stats else {}
-    stats["success"] = stats["monitor_status"] == "completed"
-
-    if stats["monitor_status"] == "error":
-        for instance, errors in stats["errors"].items():
+    if stats.monitor_status == "error":
+        err = ""
+        for instance, errors in stats.errors.items():
             for error in errors:
-                typer.secho(f"\n❌ {instance} encountered error:", fg="red", bold=True)
-                typer.secho(error, fg="red")
+                typer.secho(f"\n❌ {instance} encountered error:", fg="red", err=True, bold=True)
+                typer.secho(error, fg="red", err=True)
+                err += error + "\n"
+        UsageClient.log_exception(
+            "replicate_monitor", exceptions.SkyplaneException(err), error_reporting_args, job.source_region, job.dest_region
+        )
         raise typer.Exit(1)
-
-    # print stats
-    if stats["success"]:
-        rprint(f"\n:white_check_mark: [bold green]Transfer completed successfully[/bold green]")
-        runtime_line = f"[white]Transfer runtime:[/white] [bright_black]{stats.get('total_runtime_s'):.2f}s[/bright_black]"
-        throughput_line = f"[white]Throughput:[/white] [bright_black]{stats.get('throughput_gbits'):.2f}Gbps[/bright_black]"
-        rprint(f"{runtime_line}, {throughput_line}")
+    elif stats.monitor_status == "completed":
+        # success message will be handled by the caller
+        pass
     else:
         rprint(f"\n:x: [bold red]Transfer failed[/bold red]")
         rprint(stats)
+        UsageClient.log_exception(
+            "replicate_monitor",
+            exceptions.SkyplaneException(stats.monitor_status),
+            error_reporting_args,
+            job.source_region,
+            job.dest_region,
+        )
     return stats
