@@ -1,12 +1,16 @@
 from pathlib import Path
 
 import azure.core.exceptions
-import paramiko
+import warnings
+from cryptography.utils import CryptographyDeprecationWarning
+
+with warnings.catch_warnings():
+    warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)
+    import paramiko
 
 from skyplane import key_root
 from skyplane.compute.azure.azure_auth import AzureAuthentication
 from skyplane.compute.server import Server, ServerState
-from skyplane.utils import logger
 from skyplane.utils.cache import ignore_lru_cache
 from skyplane.utils.fn import PathLike
 
@@ -15,14 +19,7 @@ class AzureServer(Server):
     resource_group_name = "skyplane"
     resource_group_location = "westus2"
 
-    def __init__(
-        self,
-        name: str,
-        key_root: PathLike = key_root / "azure",
-        log_dir=None,
-        ssh_private_key=None,
-        assume_exists=True,
-    ):
+    def __init__(self, name: str, key_root: PathLike = key_root / "azure", log_dir=None, ssh_private_key=None, assume_exists=True):
         self.auth = AzureAuthentication()
         self.name = name
         self.location = None
@@ -149,15 +146,8 @@ class AzureServer(Server):
         compute_client = self.auth.get_compute_client()
         network_client = self.auth.get_network_client()
 
-        # remove any role assignments to the VM's system assigned identity
-        auth_client = self.auth.get_authorization_client()
-        vm = self.get_virtual_machine()
-        for assignment in auth_client.role_assignments.list(filter="principalId eq '{}'".format(vm.identity.principal_id)):
-            logger.fs.debug(f"Deleting role assignment {assignment.name}")
-            auth_client.role_assignments.delete(
-                scope=assignment.scope,
-                role_assignment_name=assignment.name,
-            )
+        self.auth.get_authorization_client()
+        self.get_virtual_machine()
 
         vm_poller = compute_client.virtual_machines.begin_delete(AzureServer.resource_group_name, self.vm_name(self.name))
         _ = vm_poller.result()
