@@ -1,20 +1,19 @@
 import json
+import math
 import pickle
 import time
 import uuid
-import math
+from dataclasses import dataclass
 from datetime import datetime
 from functools import partial
-from typing import Dict, List, Optional, Tuple, Iterable
+from typing import Any, Dict, Iterable, List, Optional, Tuple
+
 import nacl.secret
 import nacl.utils
-
 import pandas as pd
-from rich.progress import Progress, SpinnerColumn, TextColumn, TimeRemainingColumn, DownloadColumn, BarColumn, TransferSpeedColumn
 import urllib3
-
-from skyplane import GB, MB, gateway_docker_image, tmp_log_dir
-from skyplane import exceptions
+from rich.progress import BarColumn, DownloadColumn, Progress, SpinnerColumn, TextColumn, TimeRemainingColumn, TransferSpeedColumn
+from skyplane import GB, MB, exceptions, gateway_docker_image, tmp_log_dir
 from skyplane.chunk import Chunk, ChunkRequest, ChunkState
 from skyplane.compute.aws.aws_cloud_provider import AWSCloudProvider
 from skyplane.compute.azure.azure_cloud_provider import AzureCloudProvider
@@ -42,6 +41,26 @@ def refresh_instance_list(provider: CloudProvider, region_list: Iterable[str] =
     return {r: ilist for r, ilist in results if ilist}
 
 
+@dataclass
+class TransferStats:
+    monitor_status: str
+    total_runtime_s: Optional[float] = None
+    throughput_gbits: Optional[float] = None
+    errors: Optional[Dict[str, List[str]]] = None
+
+    @classmethod
+    def empty(cls):
+        return TransferStats(monitor_status="empty")
+
+    def to_dict(self) -> Dict[str, Optional[Any]]:
+        return {
+            "monitor_status": self.monitor_status,
+            "total_runtime_s": self.total_runtime_s,
+            "throughput_gbits": self.throughput_gbits,
+            "errors": [str(e) for e in self.errors.values()] if self.errors else None,
+        }
+
+
 class ReplicatorClient:
     def __init__(
         self,
@@ -84,6 +103,9 @@ class ReplicatorClient:
         use_compression=True,
         use_e2ee=True,
         use_socket_tls=False,
+        aws_use_spot_instances: bool = False,
+        azure_use_spot_instances: bool = False,
+        gcp_use_spot_instances: bool = False,
     ):
         regions_to_provision = [node.region for node in self.topology.gateway_nodes]
         aws_regions_to_provision = [r for r in regions_to_provision if r.startswith("aws:")]
@@ -112,7 +134,7 @@ class ReplicatorClient:
                     self.aws, set([r.split(":")[1] for r in aws_regions_to_provision]), aws_instance_filter
                 )
                 for r, ilist in current_aws_instances.items():
-                    for i in ilist:
+                    for _ in ilist:
                         if f"aws:{r}" in aws_regions_to_provision:
                             aws_regions_to_provision.remove(f"aws:{r}")
             else:
@@ -128,7 +150,7 @@ class ReplicatorClient:
                     self.azure, set([r.split(":")[1] for r in azure_regions_to_provision]), azure_instance_filter
                 )
                 for r, ilist in current_azure_instances.items():
-                    for i in ilist:
+                    for _ in ilist:
                         if f"azure:{r}" in azure_regions_to_provision:
                             azure_regions_to_provision.remove(f"azure:{r}")
             else:
@@ -144,7 +166,7 @@ class ReplicatorClient:
                     self.gcp, set([r.split(":")[1] for r in gcp_regions_to_provision]), gcp_instance_filter
                 )
                 for r, ilist in current_gcp_instances.items():
-                    for i in ilist:
+                    for _ in ilist:
                         if f"gcp:{r}" in gcp_regions_to_provision:
                             gcp_regions_to_provision.remove(f"gcp:{r}")
             else:
@@ -176,14 +198,19 @@ class ReplicatorClient:
             provider, subregion = region.split(":")
             if provider == "aws":
                 assert self.aws.auth.enabled()
-                server = self.aws.provision_instance(subregion, self.aws_instance_class)
+                server = self.aws.provision_instance(subregion, self.aws_instance_class, use_spot_instances=aws_use_spot_instances)
             elif provider == "azure":
                 assert self.azure.auth.enabled()
-                server = self.azure.provision_instance(subregion, self.azure_instance_class)
+                server = self.azure.provision_instance(subregion, self.azure_instance_class, use_spot_instances=azure_use_spot_instances)
             elif provider == "gcp":
                 assert self.gcp.auth.enabled()
                 # todo specify network tier in ReplicationTopology
-                server = self.gcp.provision_instance(subregion, self.gcp_instance_class, premium_network=self.gcp_use_premium_network)
+                server = self.gcp.provision_instance(
+                    subregion,
+                    self.gcp_instance_class,
+                    premium_network=self.gcp_use_premium_network,
+                    use_spot_instances=gcp_use_spot_instances,
+                )
             else:
                 raise NotImplementedError(f"Unknown provider {provider}")
             server.enable_auto_shutdown()
@@ -302,7 +329,7 @@ class ReplicatorClient:
         job: ReplicationJob,
         multipart_enabled: bool,
         multipart_min_threshold_mb: int,
-        multipart_min_size_mb: int,
+        multipart_chunk_size_mb: int,
         multipart_max_chunks: int,
     ) -> ReplicationJob:
         assert job.source_region.split(":")[0] in [
@@ -316,11 +343,7 @@ class ReplicatorClient:
             "gcp",
         ], f"Only AWS, Azure, and GCP are supported, but got {job.dest_region}"
 
-        with Progress(
-            SpinnerColumn(),
-            TextColumn("Preparing replication plan{task.description}"),
-            transient=True,
-        ) as progress:
+        with Progress(SpinnerColumn(), TextColumn("Preparing replication plan{task.description}"), transient=True) as progress:
             prepare_task = progress.add_task("", total=None)
 
             # pre-fetch instance IPs for all gateways
@@ -328,10 +351,13 @@ class ReplicatorClient:
             gateway_ips: Dict[Server, str] = {s: s.public_ip() for s in self.bound_nodes.values()}
 
             # make list of chunks
-            progress.update(prepare_task, description=": Creating list of chunks for transfer")
+            n_objs = 0
             chunks = []
+            multipart_pairs = []
             idx = 0
             for (src_object, dest_object) in job.transfer_pairs:
+                progress.update(prepare_task, description=f": Creating list of chunks for transfer ({n_objs}/{len(job.transfer_pairs)})")
+                n_objs += 1
                 if job.random_chunk_size_mb:
                     chunks.append(
                         Chunk(
@@ -344,53 +370,8 @@ class ReplicatorClient:
                     )
                     idx += 1
                 elif multipart_enabled and src_object.size > multipart_min_threshold_mb * MB:
-                    # determine number of chunks via the following algorithm:
-                    chunk_size_bytes = int(multipart_min_size_mb * MB)
-                    num_chunks = math.ceil(src_object.size / chunk_size_bytes)
-                    if num_chunks > multipart_max_chunks:
-                        chunk_size_bytes = int(src_object.size / multipart_max_chunks)
-                        chunk_size_bytes = math.ceil(chunk_size_bytes / MB) * MB  # round to next largest MB
-                        num_chunks = math.ceil(src_object.size / chunk_size_bytes)
-
-                    # TODO: potentially do this in a seperate thread, and/or after chunks sent
-                    obj_store_interface = ObjectStoreInterface.create(job.dest_region, job.dest_bucket)
-                    logger.fs.info(f"Initiate multipart upload on {dest_object}")
-                    upload_id = obj_store_interface.initiate_multipart_upload(dest_object.key)
-
-                    offset = 0
-                    part_num = 1
-                    parts = []
-                    for chunk in range(num_chunks):
-                        # size is min(chunk_size, remaining data)
-                        file_size_bytes = min(chunk_size_bytes, src_object.size - offset)
-                        assert file_size_bytes > 0, f"File size <= 0 {file_size_bytes}"
-                        chunks.append(
-                            Chunk(
-                                src_key=src_object.key,
-                                dest_key=dest_object.key,
-                                chunk_id=idx,
-                                file_offset_bytes=offset,
-                                chunk_length_bytes=file_size_bytes,
-                                part_number=part_num,
-                                upload_id=upload_id,
-                            )
-                        )
-                        parts.append(part_num)
-
-                        idx += 1
-                        part_num += 1
-                        offset += chunk_size_bytes
-                    # add multipart upload request
-                    self.multipart_upload_requests.append(
-                        {
-                            "region": job.dest_region,
-                            "bucket": job.dest_bucket,
-                            "upload_id": upload_id,
-                            "key": dest_object.key,
-                            "parts": parts,
-                        }
-                    )
-                # transfer entire object
+                    # transfer entire object
+                    multipart_pairs.append((src_object, dest_object))
                 else:
                     chunk = Chunk(
                         src_key=src_object.key,
@@ -402,6 +383,51 @@ class ReplicatorClient:
                     chunks.append(chunk)
                     idx += 1
 
+            # initiate multipart transfers in parallel
+            progress.update(prepare_task, description=f": Queuing {len(multipart_pairs)} files for multipart transfers")
+            obj_store_interface = ObjectStoreInterface.create(job.dest_region, job.dest_bucket)
+            upload_ids = do_parallel(lambda x: obj_store_interface.initiate_multipart_upload(x[1].key), multipart_pairs, n=16)
+            for (src_object, dest_object), upload_id in upload_ids:
+                # determine number of chunks via the following algorithm:
+                chunk_size_bytes = int(multipart_chunk_size_mb * MB)
+                num_chunks = math.ceil(src_object.size / chunk_size_bytes)
+                if num_chunks > multipart_max_chunks:
+                    chunk_size_bytes = int(src_object.size / multipart_max_chunks)
+                    chunk_size_bytes = math.ceil(chunk_size_bytes / MB) * MB  # round to next largest mb
+                    num_chunks = math.ceil(src_object.size / chunk_size_bytes)
+                offset = 0
+                part_num = 1
+                parts = []
+                for chunk in range(num_chunks):
+                    file_size_bytes = min(chunk_size_bytes, src_object.size - offset)  # size is min(chunk_size, remaining data)
+                    assert file_size_bytes > 0, f"file size <= 0 {file_size_bytes}"
+                    chunks.append(
+                        Chunk(
+                            src_key=src_object.key,
+                            dest_key=dest_object.key,
+                            chunk_id=idx,
+                            file_offset_bytes=offset,
+                            chunk_length_bytes=file_size_bytes,
+                            part_number=part_num,
+                            upload_id=upload_id,
+                        )
+                    )
+                    parts.append(part_num)
+
+                    idx += 1
+                    part_num += 1
+                    offset += chunk_size_bytes
+                # add multipart upload request
+                self.multipart_upload_requests.append(
+                    {
+                        "region": job.dest_region,
+                        "bucket": job.dest_bucket,
+                        "upload_id": upload_id,
+                        "key": dest_object.key,
+                        "parts": parts,
+                    }
+                )
+
             # partition chunks into roughly equal-sized batches (by bytes)
             def partition(items: List[Chunk], n_batches: int) -> List[List[Chunk]]:
                 batches = [[] for _ in range(n_batches)]
@@ -511,6 +537,7 @@ class ReplicatorClient:
         job: ReplicationJob,
         show_spinner=False,
         log_interval_s: Optional[float] = None,
+        log_to_file: bool = True,
         time_limit_seconds: Optional[float] = None,
         cleanup_gateway: bool = True,
         save_log: bool = True,
@@ -518,7 +545,7 @@ class ReplicatorClient:
         write_socket_profile: bool = False,  # slow but useful for debugging
         copy_gateway_logs: bool = False,
         multipart: bool = False,  # multipart object uploads/downloads
-    ) -> Optional[Dict]:
+    ) -> TransferStats:
         assert job.chunk_requests is not None
         total_bytes = sum([cr.chunk.chunk_length_bytes for cr in job.chunk_requests])
         last_log = None
@@ -528,8 +555,6 @@ class ReplicatorClient:
         sinks = self.topology.sink_instances()
         sink_regions = set(s.region for s in sinks)
 
-        completed_chunk_ids = []
-
         if save_log:
             (self.transfer_dir / "job.pkl").write_bytes(pickle.dumps(job))
         try:
@@ -554,16 +579,16 @@ class ReplicatorClient:
                             copy_gateway_logs = True
                             write_profile = True
                             write_socket_profile = True
-
-                            return {
-                                "errors": errors,
-                                "monitor_status": "error",
-                            }
+                            return TransferStats(
+                                monitor_status="error",
+                                total_runtime_s=t.elapsed,
+                                errors=errors,
+                            )
 
                         log_df = self.get_chunk_status_log_df()
                         if log_df.empty:
                             logger.warning("No chunk status log entries yet")
-                            time.sleep(0.5)
+                            time.sleep(0.01 if show_spinner else 0.25)
                             continue
 
                         is_complete_rec = (
@@ -606,34 +631,35 @@ class ReplicatorClient:
                                     desc="Completing multipart uploads",
                                     spinner=False,
                                 )
-                            return dict(
-                                completed_chunk_ids=completed_chunk_ids,
+                            return TransferStats(
+                                monitor_status="completed",
                                 total_runtime_s=total_runtime_s,
                                 throughput_gbits=throughput_gbits,
-                                monitor_status="completed",
                             )
                         elif time_limit_seconds is not None and t.elapsed > time_limit_seconds or t.elapsed > 600 and completed_bytes == 0:
                             logger.error("Transfer timed out without progress, please check the debug log!")
                             logger.fs.error("Transfer timed out! Please retry.")
                             logger.error(f"Please share debug logs from: {self.transfer_dir}")
-                            return dict(
-                                completed_chunk_ids=completed_chunk_ids,
+                            return TransferStats(
+                                monitor_status="timed_out",
                                 total_runtime_s=total_runtime_s,
                                 throughput_gbits=throughput_gbits,
-                                monitor_status="timed_out",
                             )
                         else:
                             current_time = datetime.now()
                             if log_interval_s and (not last_log or (current_time - last_log).seconds > float(log_interval_s)):
                                 last_log = current_time
+                                log_str = f"{total_runtime_s}s: {len(completed_chunk_ids)}/{len(job.chunk_requests)} chunks, {completed_bytes}/{total_bytes} bytes, "
+                                log_str += f"{throughput_gbits:.2f} Gbit/s"
+                                if log_to_file:
+                                    logger.fs.debug(log_str)
+                                else:
+                                    logger.debug(log_str)
                             time.sleep(0.01 if show_spinner else 0.25)
+                            continue
         # always run cleanup, even if there's an exception
         finally:
-            with Progress(
-                SpinnerColumn(),
-                TextColumn("Cleaning up after transfer{task.description}"),
-                transient=True,
-            ) as progress:
+            with Progress(SpinnerColumn(), TextColumn("Cleaning up after transfer{task.description}"), transient=True) as progress:
                 cleanup_task = progress.add_task("", total=None)
 
                 # get compression ratio information from destination gateways using "/api/v1/profile/compression"
@@ -692,6 +718,7 @@ class ReplicatorClient:
 
                     do_parallel(fn, self.bound_nodes.values(), n=-1)
                     progress.update(cleanup_task, description=": Shutting down gateways")
+        raise exceptions.SkyplaneException("Transfer failed, cleanup did not run")  # should never get here
 
     @staticmethod
     def verify_transfer_prefix(job: ReplicationJob, dest_prefix: str):
@@ -709,7 +736,4 @@ class ReplicatorClient:
                 del dst_keys[obj.key]
 
         if dst_keys:
-            raise exceptions.TransferFailedException(
-                f"{len(dst_keys)} objects failed verification",
-                [obj.key for obj in dst_keys.values()],
-            )
+            raise exceptions.TransferFailedException(f"{len(dst_keys)} objects failed verification", [obj.key for obj in dst_keys.values()])
